import inspect

from django.core.management.base import BaseCommand

from wagtail_vector_index.storage import registry


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Tells Django to NOT prompt the user for input of any kind.",
        )

    def handle(self, **options):
        interactive = options["interactive"]

        if interactive:
            confirm = input(
                inspect.cleandoc(
                    """
                WARNING:
                You are triggering a Wagtail Vector Index index update.

                Depending on your configured AI backend;
                 * Multiple API calls will be triggered to chargeable services.
                 * Contents from every indexed model will be sent to third-party APIs.

                Are you sure you want to do this?
                Type 'yes' to continue, or 'no' to cancel: """
                )
            )
        else:
            confirm = "yes"

        if confirm == "yes":
            self.stdout.write("Rebuilding vector indexes")
            for _, index in registry:
                index.rebuild_index()
        else:
            self.stdout.write("Index update cancelled")



import logging
from collections.abc import Generator, Iterable, MutableSequence, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, cast

from wagtail_vector_index.storage.base import (
    Document,
    StorageProvider,
    StorageVectorIndexMixinProtocol,
)

from .types import DistanceMethod

if TYPE_CHECKING:
    from .models import PgvectorEmbedding, PgvectorEmbeddingQuerySet

    MixinBase = StorageVectorIndexMixinProtocol["PgvectorStorageProvider"]
else:
    MixinBase = object

logger = logging.Logger(__name__)

__all__ = [
    "PgvectorStorageProvider",
    "PgvectorStorageProviderConfig",
]


def _embedding_model():
    """Lazy load the model to prevent Django trying to import it before the app registry is ready."""
    from .models import PgvectorEmbedding

    return PgvectorEmbedding


@dataclass(kw_only=True)
class PgvectorStorageProviderConfig:
    upsert_batch_size: int = 500
    bulk_create_batch_size: int = 500
    bulk_create_ignore_conflicts: bool = True
    distance_method: DistanceMethod = DistanceMethod.COSINE

    def __init__(self, **kwargs: Any) -> None:
        self.distance_method = DistanceMethod(kwargs.get("distance_method", "cosine"))
        super().__init__(**kwargs)


class PgvectorIndexMixin(MixinBase):
    upsert_batch_size: ClassVar[int] = 500
    bulk_create_batch_size: ClassVar[int] = 500
    bulk_create_ignore_conflicts: ClassVar[bool] = True
    distance_method: ClassVar[DistanceMethod] = DistanceMethod.COSINE

    def rebuild_index(self) -> None:
        self.clear()
        self.upsert(documents=self.get_documents())

    def upsert(self, *, documents: Iterable[Document]) -> None:
        counter = 0
        objs_to_create: MutableSequence["PgvectorEmbedding"] = []
        for document in documents:
            objs_to_create.append(self._document_to_embedding(document))
            if (counter + 1) % self.upsert_batch_size == 0:
                self._bulk_create(objs_to_create)
                objs_to_create = []
            counter += 1
        if objs_to_create:
            self._bulk_create(objs_to_create)

    def delete(self, *, document_ids: Sequence[str]) -> None:
        self._get_queryset().filter(embedding__pk__in=document_ids).delete()

    def clear(self):
        self._get_queryset().delete()

    def get_similar_documents(
        self, query_vector, *, limit: int = 5
    ) -> Generator[Document, None, None]:
        for pgvector_embedding in (
            self._get_queryset()
            .select_related("embedding")
            .filter(embedding_output_dimensions=len(query_vector))
            .order_by_distance(
                query_vector,
                distance_method=self.distance_method,
                fetch_distance=False,
            )[:limit]
            .iterator()
        ):
            embedding = pgvector_embedding.embedding
            yield embedding.to_document()

    def _get_queryset(self) -> "PgvectorEmbeddingQuerySet":
        # objects is technically a Manager instance but we want to use the custom
        # queryset method
        return cast("PgvectorEmbeddingQuerySet", _embedding_model().objects).in_index(
            type(self).__name__
        )

    def _bulk_create(self, embeddings: Sequence["PgvectorEmbedding"]) -> None:
        _embedding_model().objects.bulk_create(
            embeddings,
            batch_size=self.bulk_create_batch_size,
            ignore_conflicts=self.bulk_create_ignore_conflicts,
        )

    def _document_to_embedding(self, document: Document) -> "PgvectorEmbedding":
        return _embedding_model()(
            embedding_id=document.embedding_pk,
            embedding_output_dimensions=len(document.vector),
            vector=document.vector,
            index_name=type(self).__name__,
        )


class PgvectorStorageProvider(
    StorageProvider[PgvectorStorageProviderConfig, PgvectorIndexMixin]
):
    config_class = PgvectorStorageProviderConfig
    index_mixin = PgvectorIndexMixin



import json
from collections.abc import Generator, Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import weaviate
from django.core.exceptions import ImproperlyConfigured

from wagtail_vector_index.storage.base import (
    Document,
    StorageProvider,
    StorageVectorIndexMixinProtocol,
    VectorIndex,
)


@dataclass
class ProviderConfig:
    HOST: str
    API_KEY: str | None = None


if TYPE_CHECKING:
    MixinBase = StorageVectorIndexMixinProtocol["WeaviateStorageProvider"]
else:
    MixinBase = object


class WeaviateIndexMixin(MixinBase):
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self.storage_provider = self._get_storage_provider()
        self.index_name = self.__class__.__name__

    def rebuild_index(
        self,
    ) -> None:
        self.storage_provider.client.schema.delete_class(self.index_name)
        self.storage_provider.client.schema.create_class(
            {
                "class": self.index_name,
            }
        )
        self.upsert(documents=self.get_documents())

    def upsert(self, *, documents: Iterable[Document]) -> None:
        with self.storage_provider.client.batch as batch:
            for document in documents:
                # Store metadata as a JSON string because otherwise
                # we need to explicitly request each field back in
                # the query
                batch.add_data_object(
                    {
                        "metadata": json.dumps(document.metadata),
                        "embedding_pk": document.embedding_pk,
                    },
                    self.index_name,
                    vector=document.vector,
                )

    def delete(self, *, document_ids: Sequence[str]) -> None:
        # TODO: Handle deletion
        raise NotImplementedError

    def get_similar_documents(
        self, query_vector: Sequence[float], *, limit: int = 5
    ) -> Generator[Document, None, None]:
        near_vector = {
            "vector": query_vector,
        }
        similar_documents = (
            self.storage_provider.client.query.get(
                self.index_name,
                ["embedding_pk", "metadata"],
            )
            .with_additional(["distance", "vector"])
            .with_near_vector(near_vector)
            .with_limit(limit)
            .do()
        )
        docs = similar_documents["data"]["Get"][self.index_name]
        for doc in docs:
            yield Document(
                embedding_pk=doc["embedding_pk"],
                metadata=json.loads(doc["metadata"]),
                vector=doc["_additional"]["vector"],
            )


class WeaviateVectorIndex(WeaviateIndexMixin, VectorIndex):
    pass


class WeaviateStorageProvider(StorageProvider[ProviderConfig, WeaviateIndexMixin]):
    config_class = ProviderConfig
    index_mixin = WeaviateIndexMixin

    def __init__(self, config: Mapping[str, Any]) -> None:
        super().__init__(config)
        if self.config.API_KEY is None:
            raise ImproperlyConfigured("Weaviate API key is not set")
        auth_config = weaviate.auth.AuthApiKey(api_key=self.config.API_KEY)
        self.client = weaviate.Client(self.config.HOST, auth_client_secret=auth_config)



class EmbeddingOutputDimensionsNotFound(Exception):
    pass


def get_default_embedding_output_dimensions(model_id: str) -> int:
    match model_id:
        case "ada-002":
            return 1536
        case _:
            raise EmbeddingOutputDimensionsNotFound(model_id)



# Generated by Django 4.2.7 on 2023-11-22 14:46

from django.db import migrations
from pgvector.django import VectorExtension


class Migration(migrations.Migration):
    operations = [VectorExtension()]



from collections.abc import Callable
from typing import Any, Protocol, TypedDict


class ChatMessage(TypedDict):
    """Dict contaning 'role' and 'content' keys representing a single message in
    a chat conversation."""

    role: str
    content: str


class AIResponseStreamingPart(TypedDict):
    """Represents a part of a streaming AI response where the index represents the choice
    if multiple choices are returned by the backend."""

    index: int
    content: str


class AIStreamingResponse:
    """Iterator protocol representing a streaming response from an AI backend.

    Each iteration returns a single AIResponseStreamingPart. As a backend may return multiple
    'choices' in an undetermined order, the consumer must reconstruct the response based on the 'index'
    field of each part.

    e.g. This will return something something like:
    [
        {"index": 0, "content": "I"},
        {"index": 1, "content": "ChatGPT"},
        {"index": 1, "content": "is"},
        {"index": 1, "content": "an"},
        {"index": 0, "content": "am"},
        {"index": 0, "content": "ChatGPT"},
        {"index": 1, "content": "AI"},
    ]
    """

    def __iter__(self):
        return self

    def __next__(self) -> AIResponseStreamingPart: ...


class AIResponse:
    """Representation of a non-streaming response from an AI backend.

    `choices` is a list of strings representing the choices returned by the backend.
    """

    def __init__(self, choices: list[str]) -> None:
        self.choices = choices


class TextSplitterProtocol(Protocol):
    def __init__(
        self,
        *,
        chunk_size: int,
        chunk_overlap: int,
        length_function: Callable[[str], int],
        **kwargs: Any,
    ) -> None: ...

    def split_text(self, text: str) -> list[str]: ...


class TextSplitterLengthCalculatorProtocol(Protocol):
    def get_splitter_length(self, text: str) -> int: ...



from collections.abc import Mapping

from django.conf import settings

from .ai_utils.backends import BackendSettingsDict, EmbeddingBackendSettingsDict
from .ai_utils.backends import get_chat_backend as get_chat_backend_instance
from .ai_utils.backends import (
    get_embedding_backend as get_embedding_backend_instance,
)
from .ai_utils.backends.base import BaseChatBackend, BaseEmbeddingBackend


def get_chat_backends_settings() -> Mapping[str, BackendSettingsDict]:
    try:
        return settings.WAGTAIL_VECTOR_INDEX["CHAT_BACKENDS"]
    except (KeyError, AttributeError):
        return {
            "default": {
                "CLASS": "ai_utils.ai.litellm.LiteLLMChatBackend",
                "CONFIG": {
                    "MODEL_ID": "gpt-3.5-turbo",
                },
            }
        }


def get_embedding_backends_settings() -> Mapping[str, EmbeddingBackendSettingsDict]:
    try:
        return settings.WAGTAIL_VECTOR_INDEX["EMBEDDING_BACKENDS"]
    except (KeyError, AttributeError):
        return {
            "default": {
                "CLASS": "ai_utils.ai.ltiellm.LiteLLMEmbeddingBackend",
                "CONFIG": {
                    "MODEL_ID": "text-embedding-ada-002",
                },
            }
        }


def get_chat_backend(alias: str) -> BaseChatBackend:
    setting = get_chat_backends_settings()[alias]
    return get_chat_backend_instance(backend_dict=setting, backend_id=alias)


def get_embedding_backend(alias: str) -> BaseEmbeddingBackend:
    setting = get_embedding_backends_settings()[alias]
    return get_embedding_backend_instance(backend_dict=setting, backend_id=alias)



from collections.abc import Generator, Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from qdrant_client import QdrantClient
from qdrant_client.http import models as qdrant_models
from qdrant_client.models import Distance

from wagtail_vector_index.storage.base import (
    Document,
    StorageProvider,
    StorageVectorIndexMixinProtocol,
)


@dataclass
class ProviderConfig:
    HOST: str
    API_KEY: str | None = None


if TYPE_CHECKING:
    MixinBase = StorageVectorIndexMixinProtocol["QdrantStorageProvider"]
else:
    MixinBase = object


class QdrantIndexMixin(MixinBase):
    def __init__(self, **kwargs: Any) -> None:
        self.index_name = self.__class__.__name__
        self.storage_provider = self._get_storage_provider()
        super().__init__(**kwargs)

    def rebuild_index(self) -> None:
        self.storage_provider.client.delete_collection(collection_name=self.index_name)
        self.storage_provider.client.create_collection(
            collection_name=self.index_name,
            vectors_config=qdrant_models.VectorParams(
                size=512, distance=Distance.COSINE
            ),
        )
        self.upsert(documents=self.get_documents())

    def upsert(self, *, documents: Iterable[Document]) -> None:
        points = [
            qdrant_models.PointStruct(
                id=document.embedding_pk,
                vector=document.vector,
                payload=document.metadata,
            )
            for document in documents
        ]
        self.storage_provider.client.upsert(
            collection_name=self.index_name, points=points
        )

    def delete(self, *, document_ids: Sequence[str]) -> None:
        self.storage_provider.client.delete(
            collection_name=self.index_name,
            points_selector=qdrant_models.PointIdsList(points=document_ids),
        )

    def get_similar_documents(
        self, query_vector: Sequence[float], *, limit: int = 5
    ) -> Generator[Document, None, None]:
        similar_documents = self.storage_provider.client.search(
            collection_name=self.index_name, query_vector=query_vector, limit=limit
        )
        for doc in similar_documents:
            yield Document(
                embedding_pk=doc["id"], vector=doc["vector"], metadata=doc["payload"]
            )


class QdrantStorageProvider(StorageProvider[ProviderConfig, QdrantIndexMixin]):
    config_class = ProviderConfig
    index_mixin = QdrantIndexMixin

    def __init__(self, config: Mapping[str, Any]) -> None:
        super().__init__(config)
        self.client = QdrantClient(url=self.config.HOST, api_key=self.config.API_KEY)

    def rebuild_indexes(self) -> None:
        pass



"""This module contains 'Echo' backends for chat and embedding, which echo back the input.
They are intended for testing and development purposes and are not suitable for production use."
"""

import random
import time
from collections.abc import Generator, Iterable, Iterator, Sequence
from dataclasses import dataclass
from typing import Any, NotRequired, Self

from django.core.exceptions import ImproperlyConfigured

from wagtail_vector_index.storage.base import Document

from ..types import (
    AIResponse,
    AIResponseStreamingPart,
    AIStreamingResponse,
    ChatMessage,
)
from .base import (
    BaseChatBackend,
    BaseChatConfig,
    BaseChatConfigSettingsDict,
    BaseEmbeddingBackend,
    BaseEmbeddingConfig,
)


class EchoStreamingResponse(AIResponse):
    def __init__(self, response_iterator: Iterator[str]) -> None:
        self.response_iterator = response_iterator

    def __iter__(self) -> Iterator[AIResponseStreamingPart]:
        return self

    def __next__(self) -> AIResponseStreamingPart:
        return {"index": 0, "content": next(self.response_iterator)}


@dataclass(kw_only=True)
class EchoChatSettingsDict(BaseChatConfigSettingsDict):
    MAX_WORD_SLEEP_SECONDS: NotRequired[int]


@dataclass(kw_only=True)
class EchoChatConfig(BaseChatConfig[EchoChatSettingsDict]):
    max_word_sleep_seconds: int

    @classmethod
    def from_settings(cls, config: EchoChatSettingsDict, **kwargs: Any) -> Self:
        max_word_sleep_seconds = config.get("MAX_WORD_SLEEP_SECONDS")
        if max_word_sleep_seconds is None:
            max_word_sleep_seconds = 0
        try:
            max_word_sleep_seconds = int(max_word_sleep_seconds)
        except ValueError as e:
            raise ImproperlyConfigured(
                f'"MAX_WORD_SLEEP_SECONDS" is not an "int", it is a "{type(max_word_sleep_seconds)}".'
            ) from e
        kwargs.setdefault("max_word_sleep_seconds", max_word_sleep_seconds)

        return super().from_settings(config, **kwargs)


class EchoChatBackend(BaseChatBackend[EchoChatConfig]):
    config_cls = EchoChatConfig
    config: EchoChatConfig

    def build_response(self, messages: Sequence[ChatMessage]) -> Sequence[str]:
        response = ["This", "is", "an", "echo", "backend:"]
        for m in messages:
            response.extend(m["content"].split())
        return response

    def streaming_iterator(self, response: Sequence[str]) -> Generator[str, None, None]:
        for word in response:
            if (
                self.config.max_word_sleep_seconds is not None
                and self.config.max_word_sleep_seconds > 0
            ):
                time.sleep(
                    random.random()
                    * random.randint(0, self.config.max_word_sleep_seconds)
                )
            yield word

    def chat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse:
        response = self.build_response(messages)
        if stream:
            return EchoStreamingResponse(self.streaming_iterator(response))
        return AIResponse(choices=[" ".join(response)])

    async def achat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse:
        return self.chat(messages=messages, stream=stream, **kwargs)


class EchoEmbeddingBackend(BaseEmbeddingBackend[BaseEmbeddingConfig]):
    config_cls = BaseEmbeddingConfig
    config: BaseEmbeddingConfig

    def embed(self, inputs: Iterable[Document]) -> Iterator[list[float]]:
        for _ in inputs:
            yield [
                random.random() for _ in range(self.config.embedding_output_dimensions)
            ]

    async def aembed(self, inputs: Iterable[Document]) -> Iterator[list[float]]:
        return self.embed(inputs=inputs)



from collections.abc import Generator, Iterable, MutableSequence, Sequence
from typing import TYPE_CHECKING, ClassVar, Optional, TypeVar, cast

from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.db import models, transaction
from django.utils.functional import classproperty  # type: ignore
from wagtail.models import Page
from wagtail.query import PageQuerySet
from wagtail.search.index import BaseField

from wagtail_vector_index.ai_utils.backends.base import BaseEmbeddingBackend
from wagtail_vector_index.ai_utils.text_splitting.langchain import (
    LangchainRecursiveCharacterTextSplitter,
)
from wagtail_vector_index.ai_utils.text_splitting.naive import (
    NaiveTextSplitterCalculator,
)
from wagtail_vector_index.ai_utils.types import TextSplitterProtocol
from wagtail_vector_index.storage import get_storage_provider, registry
from wagtail_vector_index.storage.base import (
    Document,
    DocumentRetrievalVectorIndexMixinProtocol,
    VectorIndex,
)
from wagtail_vector_index.storage.exceptions import IndexedTypeFromDocumentError

""" Everything related to indexing Django models is in this file.

This includes:

- The Embedding Django model, which is used to store embeddings for model instances in the database
- The EmbeddableFieldsMixin, which is a mixin for Django models that lets user define which fields should be used to generate embeddings
- The EmbeddableFieldsVectorIndexMixin, which is a VectorIndex mixin that expects EmbeddableFieldsMixin models
- The EmbeddableFieldsDocumentConverter, which is a DocumentConverter that knows how to convert a model instance using the EmbeddableFieldsMixin protocol to and from a Document
"""


class Embedding(models.Model):
    """Stores an embedding for a model instance"""

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    content_type_id: int
    base_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    object_id = models.CharField(
        max_length=255,
    )
    content_object = GenericForeignKey(
        "content_type", "object_id", for_concrete_model=False
    )
    vector = models.JSONField()
    content = models.TextField()

    def __str__(self):
        return f"Embedding for {self.object_id}"

    @classmethod
    def _get_base_content_type(cls, model_or_object):
        if parents := model_or_object._meta.get_parent_list():
            return ContentType.objects.get_for_model(
                parents[-1], for_concrete_model=False
            )
        else:
            return ContentType.objects.get_for_model(
                model_or_object, for_concrete_model=False
            )

    @classmethod
    def from_instance(cls, instance: models.Model) -> "Embedding":
        """Create an Embedding instance for a model instance"""
        content_type = ContentType.objects.get_for_model(instance)
        return Embedding(
            content_type=content_type,
            base_content_type=cls._get_base_content_type(instance),
            object_id=instance.pk,
        )

    @classmethod
    def get_for_instance(cls, instance: models.Model):
        """Get all Embedding instances that are related to a model instance"""
        content_type = ContentType.objects.get_for_model(instance)
        return Embedding.objects.filter(
            content_type=content_type, object_id=instance.pk
        )

    def to_document(self) -> Document:
        return Document(
            vector=self.vector,
            embedding_pk=self.pk,
            metadata={
                "object_id": str(self.object_id),
                "content_type_id": str(self.content_type_id),
                "content": self.content,
            },
        )


# ###########
# Classes that allow users to automatically generate documents from their models based on fields specified
# ###########


class EmbeddingField(BaseField):
    """A field that can be used to specify which fields of a model should be used to generate embeddings"""

    def __init__(self, *args, important=False, **kwargs):
        self.important = important
        super().__init__(*args, **kwargs)


class EmbeddableFieldsMixin(models.Model):
    """Mixin for Django models that allows the user to specify which fields should be used to generate embeddings."""

    embedding_fields = []
    embeddings = GenericRelation(
        Embedding, content_type_field="content_type", for_concrete_model=False
    )

    class Meta:
        abstract = True

    @classmethod
    def _get_embedding_fields(cls) -> list["EmbeddingField"]:
        embedding_fields = {
            (type(field), field.field_name): field for field in cls.embedding_fields
        }
        return list(embedding_fields.values())

    @classmethod
    def check(cls, **kwargs):
        """Extend model checks to include validation of embedding_fields in the
        same way that Wagtail's Indexed class does it."""
        errors = super().check(**kwargs)
        errors.extend(cls._check_embedding_fields(**kwargs))
        return errors

    @classmethod
    def _has_field(cls, name):
        try:
            cls._meta.get_field(name)
        except FieldDoesNotExist:
            return hasattr(cls, name)
        else:
            return True

    @classmethod
    def _check_embedding_fields(cls, **kwargs):
        errors = []
        for field in cls._get_embedding_fields():
            message = "{model}.embedding_fields contains non-existent field '{name}'"
            if not cls._has_field(field.field_name):
                errors.append(
                    checks.Warning(
                        message.format(model=cls.__name__, name=field.field_name),
                        obj=cls,
                        id="wagtailai.WA001",
                    )
                )
        return errors


IndexedType = TypeVar("IndexedType")


class EmbeddableFieldsDocumentConverter:
    """Implementation of DocumentConverter that knows how to convert a model instance using the
    EmbeddableFieldsMixin to and from a Document.

    Stores and retrieves embeddings from an Embedding model."""

    def __init__(self, base_model: type[models.Model]):
        # The model that this converter will convert Documents back to
        self.base_model = base_model

    def _get_split_content(
        self, object: EmbeddableFieldsMixin, *, chunk_size: int
    ) -> list[str]:
        """Split the contents of a model instance's `embedding_fields` in to smaller chunks"""
        splittable_content = []
        important_content = []
        embedding_fields = object._meta.model._get_embedding_fields()

        for field in embedding_fields:
            value = field.get_value(object)
            if value is None:
                continue
            if isinstance(value, str):
                final_value = value
            else:
                final_value: str = "\n".join((str(v) for v in value))
            if field.important:
                important_content.append(final_value)
            else:
                splittable_content.append(final_value)

        text = "\n".join(splittable_content)
        important_text = "\n".join(important_content)
        splitter = self._get_text_splitter_class(chunk_size=chunk_size)
        return [f"{important_text}\n{text}" for text in splitter.split_text(text)]

    def _get_text_splitter_class(self, chunk_size: int) -> TextSplitterProtocol:
        length_calculator = NaiveTextSplitterCalculator()
        return LangchainRecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=100,
            length_function=length_calculator.get_splitter_length,
        )

    def _existing_embeddings_match(
        self, embeddings: Iterable[Embedding], splits: list[str]
    ) -> bool:
        """Determine whether the embeddings passed in match the text content passed in"""
        if not embeddings:
            return False

        embedding_content = {embedding.content for embedding in embeddings}

        return set(splits) == embedding_content

    @transaction.atomic
    def generate_embeddings(
        self, object: EmbeddableFieldsMixin, *, embedding_backend: BaseEmbeddingBackend
    ) -> list[Embedding]:
        """Use the AI backend to generate and store embeddings for this object"""
        splits = self._get_split_content(
            object, chunk_size=embedding_backend.config.token_limit
        )
        embeddings = Embedding.get_for_instance(object)

        # If the existing embeddings all match on content, we return them
        # without generating new ones
        if self._existing_embeddings_match(embeddings, splits):
            return list(embeddings)

        # Otherwise we delete all the existing embeddings and get new ones
        embeddings.delete()

        embedding_vectors = embedding_backend.embed(splits)
        generated_embeddings: MutableSequence[Embedding] = []
        for idx, returned_embedding in enumerate(embedding_vectors):
            split = splits[idx]
            embedding = Embedding.from_instance(object)
            embedding.vector = returned_embedding
            embedding.content = split
            embedding.save()
            generated_embeddings.append(embedding)

        return generated_embeddings

    def to_documents(
        self, object: EmbeddableFieldsMixin, *, embedding_backend: BaseEmbeddingBackend
    ) -> Generator[Document, None, None]:
        for embedding in self.generate_embeddings(
            object, embedding_backend=embedding_backend
        ):
            yield embedding.to_document()

    def from_document(self, document: Document) -> models.Model:
        if obj := self.base_model.objects.filter(
            pk=document.metadata["object_id"],
            content_type=document.metadata["content_type_id"],
        ).first():
            return obj
        else:
            raise IndexedTypeFromDocumentError("No object found for document")

    def bulk_to_documents(
        self,
        objects: Iterable[EmbeddableFieldsMixin],
        *,
        embedding_backend: BaseEmbeddingBackend,
    ) -> Generator[Document, None, None]:
        # TODO: Implement a more efficient bulk embedding approach
        for object in objects:
            yield from self.to_documents(object, embedding_backend=embedding_backend)

    def bulk_from_documents(
        self, documents: Iterable[Document]
    ) -> Generator[models.Model, None, None]:
        # TODO: Implement a more efficient approach
        for document in documents:
            yield self.from_document(document)


# ###########
# VectorIndex mixins which add model-specific behaviour
# ###########

if TYPE_CHECKING:
    MixinBase = DocumentRetrievalVectorIndexMixinProtocol
else:
    MixinBase = object


class EmbeddableFieldsVectorIndexMixin(MixinBase):
    """A Mixin for VectorIndex which indexes the results of querysets of EmbeddableFieldsMixin models"""

    querysets: ClassVar[Sequence[models.QuerySet]]

    def _get_querysets(self) -> Sequence[models.QuerySet]:
        return self.querysets

    def get_converter_class(self) -> type[EmbeddableFieldsDocumentConverter]:
        return EmbeddableFieldsDocumentConverter

    def get_converter(self) -> EmbeddableFieldsDocumentConverter:
        queryset_models = [qs.model for qs in self._get_querysets()]
        all_the_same = len(set(queryset_models)) == 1
        if not all_the_same:
            raise ValueError(
                "All querysets must be of the same model to use the default converter."
            )
        return self.get_converter_class()(queryset_models[0])

    def get_documents(self) -> Iterable[Document]:
        querysets = self._get_querysets()
        all_documents = []

        for queryset in querysets:
            instances = queryset.prefetch_related("embeddings")
            # We need to consume the generator here to ensure that the
            # Embedding models are created, even if it is not consumed
            # by the caller
            all_documents += list(
                self.get_converter().bulk_to_documents(
                    instances, embedding_backend=self.get_embedding_backend()
                )
            )
        return all_documents


class PageEmbeddableFieldsVectorIndexMixin(EmbeddableFieldsVectorIndexMixin):
    """A mixin for VectorIndex for use with Wagtail pages that automatically
    restricts indexed models to live pages."""

    querysets: Sequence[PageQuerySet]

    def _get_querysets(self) -> list[PageQuerySet]:
        qs_list = super()._get_querysets()

        # Technically a manager instance, not a queryset, but we want to use the custom
        # methods.
        return [cast(PageQuerySet, qs).live() for qs in qs_list]


# ###########
# Classes related to automatic generation of indexes for models
# ###########


def camel_case(snake_str: str):
    """Convert a snake_case string to CamelCase"""
    parts = snake_str.split("_")
    return "".join(*map(str.title, parts))


def build_vector_index_base_for_storage_provider(
    storage_provider_alias: str = "default",
):
    """Build a VectorIndex base class for a given storage provider alias.

    e.g. If WAGATAIL_VECTOR_INDEX_STORAGE_PROVIDERS includes a provider with alias "default" referencing the PgvectorStorageProvider,
    this function will return a class that is a subclass of PgvectorIndexMixin and VectorIndex."""

    storage_provider = get_storage_provider(storage_provider_alias)
    alias_camel = camel_case(storage_provider_alias)
    return type(
        f"{alias_camel}VectorIndex", (storage_provider.index_mixin, VectorIndex), {}
    )


# A VectorIndex built from whatever mixin belongs to the storage provider with the "default" alias
DefaultStorageVectorIndex = build_vector_index_base_for_storage_provider("default")


class GeneratedIndexMixin(models.Model):
    """Mixin for Django models that automatically generates and registers a VectorIndex for the model.

    The model can still have custom VectorIndex classes registered if needed."""

    vector_index_class: ClassVar[Optional[type[VectorIndex]]] = None

    class Meta:
        abstract = True

    @classmethod
    def generated_index_class_name(cls):
        """Return the class name to be used for the index generated by this mixin"""
        return f"{cls.__name__}Index"

    @classmethod
    def build_vector_index(cls) -> VectorIndex:
        """Build a VectorIndex instance for this model"""

        class_list = ()
        # If the user has specified a custom `vector_index_class`, use that
        if cls.vector_index_class:
            class_list = (cls.vector_index_class,)
        else:
            storage_provider = get_storage_provider("default")
            base_cls = VectorIndex
            storage_mixin_cls = storage_provider.index_mixin
            # If the model is a Wagtail Page, use a special PageEmbeddableFieldsVectorIndexMixin
            if issubclass(cls, Page):
                mixin_cls = PageEmbeddableFieldsVectorIndexMixin
            # Otherwise use the standard EmbeddableFieldsVectorIndexMixin
            else:
                mixin_cls = EmbeddableFieldsVectorIndexMixin
            class_list = (
                mixin_cls,
                storage_mixin_cls,
                base_cls,
            )

        return cast(
            type[VectorIndex],
            type(
                cls.generated_index_class_name(),
                class_list,
                {
                    "querysets": [cls.objects.all()],
                },
            ),
        )()

    @classproperty
    def vector_index(cls):
        """Get a vector index instance for this model"""

        return registry[cls.generated_index_class_name()]


class VectorIndexedMixin(EmbeddableFieldsMixin, GeneratedIndexMixin, models.Model):
    """Model mixin which adds both the embeddable fields behaviour and the automatic index behaviour to a model."""

    class Meta:
        abstract = True


def register_indexed_models():
    """Discover and register all models that are a subclass of GeneratedIndexMixin."""
    indexed_models = [
        model
        for model in apps.get_models()
        if issubclass(model, GeneratedIndexMixin) and not model._meta.abstract
    ]
    for model in indexed_models:
        registry.register_index(model.build_vector_index())



from collections.abc import Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, NotRequired, Self

import litellm
from django.core.exceptions import ImproperlyConfigured

from ..types import (
    AIResponse,
    AIResponseStreamingPart,
    AIStreamingResponse,
    ChatMessage,
)
from .base import (
    BaseChatBackend,
    BaseChatConfig,
    BaseChatConfigSettingsDict,
    BaseConfigSettingsDict,
    BaseEmbeddingBackend,
    BaseEmbeddingConfig,
    BaseEmbeddingConfigSettingsDict,
)


class BaseLiteLLMSettingsDict(BaseConfigSettingsDict):
    DEFAULT_PARAMETERS: NotRequired[Mapping[str, Any] | None]


class LiteLLMBackendSettingsDict(BaseLiteLLMSettingsDict, BaseChatConfigSettingsDict):
    pass


class LiteLLMEmbeddingSettingsDict(
    BaseLiteLLMSettingsDict, BaseEmbeddingConfigSettingsDict
):
    pass


def build_ai_response(response):
    """Convert a LiteLLM response to the appropriate AIResponse class"""

    if type(response) == litellm.CustomStreamWrapper:
        return LiteLLMStreamingAIResponse(response)

    return AIResponse(
        choices=[choice["message"]["content"] for choice in response.choices]
    )


class LiteLLMStreamingAIResponse(AIStreamingResponse):
    """A wrapper around a litellm.CustomStreamWrapper to make it compatible with the AIStreamingResponse interface."""

    def __init__(self, stream_wrapper: litellm.CustomStreamWrapper) -> None:
        self.stream_wrapper = stream_wrapper

    def __iter__(self):
        return self

    def __next__(self) -> AIResponseStreamingPart:
        next_response = next(self.stream_wrapper)
        index = next_response.choices[0].index
        choice = next_response.choices[0]
        assert isinstance(choice, litellm.utils.StreamingChoices)  # type: ignore
        content = choice.delta.content
        assert isinstance(content, str)

        return {
            "index": index,
            "content": content,
        }


@dataclass(kw_only=True)
class LiteLLMBackendConfigMixin:
    default_parameters: Mapping[str, Any]

    @classmethod
    def from_settings(cls, config: BaseLiteLLMSettingsDict, **kwargs: Any) -> Self:
        default_parameters = config.get("DEFAULT_PARAMETERS")
        if default_parameters is None:
            default_parameters = {}
        kwargs.setdefault("default_parameters", default_parameters)

        return super().from_settings(config, **kwargs)  # type: ignore

    @classmethod
    def _get_token_limit(cls, *, model_id: str) -> int:
        """Backend-specific method for retrieving the token limit for the provided model."""
        model_info = litellm.get_model_info(model=model_id)  # type: ignore
        if (
            not model_info
            or "max_input_tokens" not in model_info
            or not model_info["max_input_tokens"]
        ):
            raise ImproperlyConfigured(
                f"LiteLLM doesn't know about model {model_id}. Set `TOKEN_LIMIT` to specify the maximum tokens accepted by this model as input."
            )
        return model_info["max_input_tokens"]


@dataclass(kw_only=True)
class LiteLLMChatBackendConfig(
    LiteLLMBackendConfigMixin, BaseChatConfig[LiteLLMBackendSettingsDict]
):
    pass


@dataclass(kw_only=True)
class LiteLLMEmbeddingBackendConfig(
    LiteLLMBackendConfigMixin, BaseEmbeddingConfig[LiteLLMEmbeddingSettingsDict]
):
    @classmethod
    def _get_embedding_output_dimensions(cls, *, model_id: str) -> int:
        model_info = litellm.get_model_info(model=model_id)  # type: ignore
        if (
            not model_info
            or "output_vector_size" not in model_info
            or not model_info["output_vector_size"]
        ):
            raise ImproperlyConfigured(
                f"LiteLLM doesn't know about model {model_id}. Set `EMBEDDING_OUTPUT_DIMENSIONS` to specify the size of the embeddings generated by this model."
            )
        return model_info["output_vector_size"]


class LiteLLMChatBackend(BaseChatBackend[LiteLLMChatBackendConfig]):
    config: LiteLLMChatBackendConfig
    config_cls = LiteLLMChatBackendConfig

    def chat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse:
        parameters = {**self.config.default_parameters, **kwargs}
        response = litellm.completion(
            model=self.config.model_id,
            messages=list(messages),
            stream=stream,
            **parameters,
        )
        return build_ai_response(response)

    async def achat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse:
        parameters = {**self.config.default_parameters, **kwargs}
        response = await litellm.acompletion(
            model=self.config.model_id,
            messages=list(messages),
            stream=stream,
            **parameters,
        )
        return build_ai_response(response)


class LiteLLMEmbeddingBackend(BaseEmbeddingBackend[LiteLLMEmbeddingBackendConfig]):
    config: LiteLLMEmbeddingBackendConfig
    config_cls = LiteLLMEmbeddingBackendConfig

    def embed(self, inputs: Iterable[str], **kwargs) -> Iterator[list[float]]:
        response = litellm.embedding(model=self.config.model_id, input=inputs, **kwargs)
        # LiteLLM *should* return an EmbeddingResponse
        assert isinstance(response, litellm.EmbeddingResponse)
        yield from [data["embedding"] for data in response["data"]]

    async def aembed(self, inputs: Iterable[str], **kwargs) -> list[float]:
        response = await litellm.aembedding(
            model=self.config.model_id, inputs=inputs, **kwargs
        )

        return [data["embedding"] for data in response["data"]]



# Generated by Django 4.2.7 on 2023-12-20 12:07

import django.db.models.deletion
import pgvector.django
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("wagtail_vector_index", "0001_initial"),
        ("pgvector", "0001_enable_pgvector_extension"),
    ]

    operations = [
        migrations.CreateModel(
            name="PgvectorEmbedding",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("vector", pgvector.django.VectorField()),
                (
                    "embedding_output_dimensions",
                    models.PositiveIntegerField(db_index=True),
                ),
                ("index_name", models.TextField()),
                (
                    "embedding",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="wagtail_vector_index.embedding",
                    ),
                ),
            ],
            options={
                "verbose_name": "pgvector embedding",
                "indexes": [
                    models.Index(
                        fields=["index_name"], name="pgvector_pg_index_n_9af56b_idx"
                    )
                ],
            },
        ),
        migrations.AddConstraint(
            model_name="pgvectorembedding",
            constraint=models.UniqueConstraint(
                fields=("embedding", "index_name", "embedding_output_dimensions"),
                name="unique_pgvector_embedding_per_index_and_dimensions",
            ),
        ),
    ]



import copy
from collections.abc import Generator, Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, ClassVar, Generic, Protocol, TypeVar

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from wagtail_vector_index.ai import get_chat_backend, get_embedding_backend
from wagtail_vector_index.ai_utils.backends.base import BaseEmbeddingBackend
from wagtail_vector_index.storage import (
    get_storage_provider,
)

StorageProviderClass = TypeVar("StorageProviderClass")
ConfigClass = TypeVar("ConfigClass")
IndexMixin = TypeVar("IndexMixin")


class DocumentRetrievalVectorIndexMixinProtocol(Protocol):
    """Protocol which defines the minimum requirements for a VectorIndex to be used with a mixin that provides
    document retrieval/generation"""

    def get_embedding_backend(self) -> BaseEmbeddingBackend: ...


class StorageVectorIndexMixinProtocol(Protocol[StorageProviderClass]):
    """Protocol which defines the minimum requirements for a VectorIndex to be used with a StorageProvider mixin."""

    storage_provider: StorageProviderClass

    def rebuild_index(self) -> None: ...

    def upsert(self) -> None: ...

    def get_documents(self) -> Iterable["Document"]: ...

    def _get_storage_provider(self) -> StorageProviderClass: ...


class StorageProvider(Generic[ConfigClass, IndexMixin]):
    """Base class for a storage provider that provides methods for interacting with a provider,
    e.g. creating and managing indexes."""

    config: ConfigClass
    config_class: type[ConfigClass]
    index_mixin: type[IndexMixin]

    def __init__(self, config: Mapping[str, Any]) -> None:
        try:
            config = dict(copy.deepcopy(config))
            self.config = self.config_class(**config)
        except TypeError as e:
            raise ImproperlyConfigured(
                f"Missing configuration settings for the vector backend: {e}"
            ) from e

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if not hasattr(cls, "config_class"):
            raise AttributeError(
                f"Storage provider {cls.__name__} must specify a `config_class` class \
                    attribute"
            )
        return super().__init_subclass__(**kwargs)


@dataclass(kw_only=True, frozen=True)
class Document:
    """Representation of some content that is passed to vector storage backends.

    A document is usually a part of a model, e.g. some content split out from
    a VectorIndexedMixin model. One model instance may have multiple documents.

    The embedding_pk on a Document must be the PK of an Embedding model instance.
    """

    vector: Sequence[float]
    embedding_pk: int
    metadata: Mapping


class DocumentConverter(Protocol):
    def to_documents(
        self, object: object, *, embedding_backend: BaseEmbeddingBackend
    ) -> Generator[Document, None, None]: ...

    def from_document(self, document: Document) -> object: ...

    def bulk_to_documents(
        self, objects: Iterable[object], *, embedding_backend: BaseEmbeddingBackend
    ) -> Generator[Document, None, None]: ...

    def bulk_from_documents(
        self, documents: Iterable[Document]
    ) -> Generator[object, None, None]: ...


@dataclass
class QueryResponse:
    """Represents a response to the VectorIndex `query` method,
    including a response string and a list of sources that were used to generate the response
    """

    response: str
    sources: Iterable[object]


class VectorIndex(Generic[ConfigClass]):
    """Base class for a VectorIndex, representing some set of documents that can be queried"""

    # The alias of the backend to use for generating embeddings when documents are added to this index
    embedding_backend_alias: ClassVar[str] = "default"

    # The alias of the storage provider specified in WAGTAIL_VECTOR_INDEX_STORAGE_PROVIDERS
    storage_provider_alias: ClassVar[str] = "default"

    def get_embedding_backend(self) -> BaseEmbeddingBackend:
        return get_embedding_backend(self.embedding_backend_alias)

    def get_documents(self) -> Iterable[Document]:
        raise NotImplementedError

    def get_converter(self) -> DocumentConverter:
        raise NotImplementedError

    # Public API

    def query(
        self, query: str, *, sources_limit: int = 5, chat_backend_alias: str = "default"
    ) -> QueryResponse:
        """Perform a natural language query against the index, returning a QueryResponse containing the natural language response, and a list of sources"""
        try:
            query_embedding = next(self.get_embedding_backend().embed([query]))
        except StopIteration as e:
            raise ValueError("No embeddings were generated for the given query.") from e

        similar_documents = list(self.get_similar_documents(query_embedding))

        sources = self._deduplicate_list(
            self.get_converter().bulk_from_documents(similar_documents)
        )

        merged_context = "\n".join(doc.metadata["content"] for doc in similar_documents)
        prompt = (
            getattr(settings, "WAGTAIL_VECTOR_INDEX_QUERY_PROMPT", None)
            or "You are a helpful assistant. Use the following context to answer the question. Don't mention the context in your answer."
        )
        messages = [
            {"content": prompt, "role": "system"},
            {"content": merged_context, "role": "system"},
            {"content": query, "role": "user"},
        ]
        chat_backend = get_chat_backend(chat_backend_alias)
        response = chat_backend.chat(messages=messages)
        return QueryResponse(response=response.choices[0], sources=sources)

    def find_similar(
        self, object, *, include_self: bool = False, limit: int = 5
    ) -> list:
        """Find similar objects to the given object"""
        converter = self.get_converter()
        object_documents: Generator[Document, None, None] = converter.to_documents(
            object, embedding_backend=self.get_embedding_backend()
        )
        similar_documents = []
        for document in object_documents:
            similar_documents += self.get_similar_documents(
                document.vector, limit=limit
            )

        return self._deduplicate_list(
            converter.bulk_from_documents(similar_documents),
            exclusions=None if include_self else [object],
        )

    def search(self, query: str, *, limit: int = 5) -> list:
        """Perform a search against the index, returning only a list of matching sources"""
        try:
            query_embedding = next(self.get_embedding_backend().embed([query]))
        except StopIteration as e:
            raise ValueError("No embeddings were generated for the given query.") from e
        similar_documents = self.get_similar_documents(query_embedding, limit=limit)

        # Eliminate duplicates of the same objects.
        return self._deduplicate_list(
            self.get_converter().bulk_from_documents(similar_documents)
        )

    # Utilities

    def _get_storage_provider(self):
        provider = get_storage_provider(self.storage_provider_alias)
        if not issubclass(self.__class__, provider.index_mixin):
            raise TypeError(
                f"The storage provider with alias '{self.storage_provider_alias}' requires an index that uses the '{provider.index_mixin.__class__.__name__}' mixin."
            )
        return provider

    @staticmethod
    def _deduplicate_list(
        objects: Iterable[object],
        *,
        exclusions: Iterable[object] | None = None,
    ) -> list[object]:
        if exclusions is None:
            exclusions = []
        # This code assumes that dict.fromkeys preserves order which is
        # behavior of the Python language since version 3.7.
        return list(dict.fromkeys(item for item in objects if item not in exclusions))

    # Backend-specific methods

    def rebuild_index(self) -> None:
        raise NotImplementedError

    def upsert(self, *, documents: Iterable["Document"]) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    def delete(self, *, document_ids: Sequence[str]) -> None:
        raise NotImplementedError

    def get_similar_documents(
        self, query_vector: Sequence[float], *, limit: int = 5
    ) -> Generator[Document, None, None]:
        raise NotImplementedError



from collections.abc import Sequence
from typing import Self

from django.db import models
from django.utils.translation import gettext_lazy as _
from pgvector.django import CosineDistance, L2Distance, MaxInnerProduct, VectorField

from .types import DistanceMethod


class PgvectorEmbeddingQuerySet(models.QuerySet["PgvectorEmbedding"]):
    def in_index(self, index_name: str) -> Self:
        return self.filter(index_name=index_name)

    def _distance_method_cls(
        self, distance_method: DistanceMethod | str
    ) -> type[models.Func]:
        distance_method = DistanceMethod(distance_method)

        if distance_method == DistanceMethod.COSINE:
            return CosineDistance
        elif distance_method == DistanceMethod.EUCLIDEAN:
            return L2Distance
        elif distance_method == DistanceMethod.MAX_INNER_PRODUCT:
            return MaxInnerProduct
        else:
            raise ValueError(f"Unknown distance method: {distance_method}")

    def annotate_with_distance(
        self,
        query_vector: Sequence[float],
        *,
        distance_method: DistanceMethod | str,
        fetch_distance: bool,
    ) -> Self:
        kwargs = {
            "distance": self._distance_method_cls(distance_method)(
                "vector", query_vector
            )
        }
        if fetch_distance:
            return self.annotate(**kwargs)
        return self.alias(**kwargs)

    def order_by_distance(
        self,
        query_vector: Sequence[float],
        *,
        asc: bool = True,
        distance_method: DistanceMethod | str,
        fetch_distance: bool,
    ) -> Self:
        qs = self.annotate_with_distance(
            query_vector, distance_method=distance_method, fetch_distance=fetch_distance
        ).order_by("distance")
        if not asc:
            qs = qs.reverse()
        return qs


class PgvectorEmbeddingManager(models.Manager.from_queryset(PgvectorEmbeddingQuerySet)):
    pass


class PgvectorEmbedding(models.Model):
    embedding = models.ForeignKey(
        "wagtail_vector_index.Embedding", on_delete=models.CASCADE, related_name="+"
    )
    vector = VectorField()
    embedding_output_dimensions = models.PositiveIntegerField(db_index=True)
    index_name = models.TextField()

    objects = PgvectorEmbeddingManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["embedding", "index_name", "embedding_output_dimensions"],
                name="unique_pgvector_embedding_per_index_and_dimensions",
            )
        ]
        indexes = [
            models.Index(fields=["index_name"]),
        ]
        verbose_name = _("pgvector embedding")
        # TODO: Determine if we need to add an indexes for the vector field.
        #       https://github.com/pgvector/pgvector-python/tree/master#django

    def __str__(self) -> str:
        return "pgvector embedding for {}".format(self.embedding)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class PgvectorConfig(AppConfig):
    name = "wagtail_vector_index.storage.pgvector"
    verbose_name = _("pgvector")
    default_auto_field = "django.db.models.AutoField"

    # TODO: Check if we're using a Postgres database.
    #       Otherwise this app should not be allowed.



from collections.abc import Iterable, Iterator, Sequence
from dataclasses import dataclass
from typing import (
    Any,
    ClassVar,
    Generic,
    Literal,
    NotRequired,
    Protocol,
    Required,
    Self,
    TypedDict,
    TypeVar,
    overload,
)

from django.core.exceptions import ImproperlyConfigured

from .. import embeddings, tokens
from ..types import (
    AIResponse,
    AIStreamingResponse,
    ChatMessage,
)


class BaseConfigSettingsDict(TypedDict):
    MODEL_ID: Required[str]
    TOKEN_LIMIT: NotRequired[int | None]
    CHUNK_OVERLAP_CHARACTERS: NotRequired[int | None]


class BaseChatConfigSettingsDict(BaseConfigSettingsDict):
    pass


class BaseEmbeddingConfigSettingsDict(BaseConfigSettingsDict):
    EMBEDDING_OUTPUT_DIMENSIONS: NotRequired[int | None]


ConfigSettings = TypeVar(
    "ConfigSettings", bound=BaseConfigSettingsDict, contravariant=True
)


ChatConfigSettings = TypeVar(
    "ChatConfigSettings", bound=BaseChatConfigSettingsDict, contravariant=True
)

EmbeddingConfigSettings = TypeVar(
    "EmbeddingConfigSettings", bound=BaseEmbeddingConfigSettingsDict, contravariant=True
)


class ConfigClassProtocol(Protocol[ConfigSettings]):
    @classmethod
    def from_settings(cls, config: ConfigSettings, **kwargs: Any) -> Self: ...


@dataclass(kw_only=True)
class BaseConfig(ConfigClassProtocol[ConfigSettings]):
    model_id: str
    token_limit: int

    @classmethod
    def from_settings(
        cls,
        config: ConfigSettings,
        **kwargs: Any,
    ) -> Self:
        token_limit = cls.get_token_limit(
            model_id=config["MODEL_ID"], custom_value=config.get("TOKEN_LIMIT")
        )

        return cls(
            model_id=config["MODEL_ID"],
            token_limit=token_limit,
            **kwargs,
        )

    @classmethod
    def _get_token_limit(cls, *, model_id: str) -> int:
        """Backend-specific method for retriving the token limit for the provided model."""
        try:
            return tokens.get_default_token_limit(model_id=model_id)
        except tokens.NoTokenLimitFound as e:
            raise ImproperlyConfigured(
                f'"TOKEN_LIMIT" is not configured for model "{model_id}".'
            ) from e

    @classmethod
    def get_token_limit(cls, *, model_id: str, custom_value: int | None) -> int:
        """Determine a token limit either from the config, or from the backend-specific
        method."""
        if custom_value is not None:
            try:
                return int(custom_value)
            except ValueError as e:
                raise ImproperlyConfigured(
                    f'"TOKEN_LIMIT" is not an "int", it is a "{type(custom_value)}".'
                ) from e
        return cls._get_token_limit(model_id=model_id)


@dataclass(kw_only=True)
class BaseChatConfig(BaseConfig[ChatConfigSettings]):
    pass


@dataclass(kw_only=True)
class BaseEmbeddingConfig(BaseConfig[EmbeddingConfigSettings]):
    embedding_output_dimensions: int

    @classmethod
    def from_settings(
        cls,
        config: EmbeddingConfigSettings,
        **kwargs: Any,
    ) -> Self:
        embedding_output_dimensions = cls.get_embedding_output_dimensions(
            model_id=config["MODEL_ID"],
            custom_value=config.get("EMBEDDING_OUTPUT_DIMENSIONS"),
        )
        kwargs.setdefault("embedding_output_dimensions", embedding_output_dimensions)
        return super().from_settings(
            config=config,
            **kwargs,
        )

    @classmethod
    def _get_embedding_output_dimensions(cls, *, model_id: str) -> int:
        """Backend-specific method for retriving the embedding output dimensions for the provided model."""
        try:
            return embeddings.get_default_embedding_output_dimensions(model_id=model_id)
        except embeddings.EmbeddingOutputDimensionsNotFound as e:
            raise ImproperlyConfigured(
                f'"EMBEDDING_OUTPUT_DIMENSIONS" is not configured for model "{model_id}".'
            ) from e

    @classmethod
    def get_embedding_output_dimensions(
        cls, *, model_id: str, custom_value: int | None
    ) -> int:
        """Determine the embedding output dimensons either from the config, or from the backend-specific
        method."""
        if custom_value is not None:
            try:
                return int(custom_value)
            except ValueError as e:
                raise ImproperlyConfigured(
                    f'"EMBEDDING_OUTPUT_DIMENSIONS" is not an "int", it is a "{type(custom_value)}".'
                ) from e
        return cls._get_embedding_output_dimensions(model_id=model_id)


AnyBackendConfig = TypeVar("AnyBackendConfig", bound=BaseConfig)
ChatBackendConfig = TypeVar("ChatBackendConfig", bound=BaseChatConfig)
EmbeddingBackendConfig = TypeVar("EmbeddingBackendConfig", bound=BaseEmbeddingConfig)


class BaseBackend(Generic[AnyBackendConfig]):
    config_cls: ClassVar[type[BaseConfig]]
    config: AnyBackendConfig

    def __init__(self, *, config: AnyBackendConfig) -> None:
        self.config = config


class BaseChatBackend(BaseBackend[ChatBackendConfig]):
    """Base chat backend providing interface with `chat` and `achat` methods.

    These both return either an `AIResponse` or an `AIStreamingResponse` depending on the `stream` parameter.
    """

    config_cls: ClassVar[type[BaseChatConfig]]
    config: ChatBackendConfig

    @overload
    def chat(
        self, *, messages: Sequence[ChatMessage], stream: Literal[True], **kwargs
    ) -> AIStreamingResponse: ...

    @overload
    def chat(
        self,
        *,
        messages: Sequence[ChatMessage],
        stream: Literal[False] = False,
        **kwargs,
    ) -> AIResponse: ...

    def chat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse: ...

    @overload
    async def achat(
        self, *, messages: Sequence[ChatMessage], stream: Literal[True], **kwargs
    ) -> AIStreamingResponse: ...

    @overload
    async def achat(
        self,
        *,
        messages: Sequence[ChatMessage],
        stream: Literal[False] = False,
        **kwargs,
    ) -> AIResponse: ...

    async def achat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse | AIStreamingResponse:
        raise NotImplementedError("Async chat is not supported by this backend.")


class BaseEmbeddingBackend(BaseBackend[EmbeddingBackendConfig]):
    """Base embedding backend providing interface with `embed` and async `aembed` methods."""

    config_cls: ClassVar[type[BaseEmbeddingConfig]]
    config: EmbeddingBackendConfig

    def embed(self, inputs: Iterable[str], **kwargs) -> Iterator[list[float]]: ...

    async def aembed(self, inputs: Iterable[str]) -> Iterator[list[float]]:
        raise NotImplementedError("Async embed is not supported by this backend.")

    @property
    def embedding_output_dimensions(self) -> int:
        return self.config.embedding_output_dimensions



from django.apps import AppConfig


class WagtailVectorIndexAppConfig(AppConfig):
    label = "wagtail_vector_index"
    name = "wagtail_vector_index"
    verbose_name = "Wagtail Vector Index"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from wagtail_vector_index.storage.models import register_indexed_models

        register_indexed_models()



import os
from collections.abc import Iterable, Iterator, Mapping, Sequence
from typing import Any, NotRequired, Self

import llm
from llm.models import dataclass

from ..types import AIResponse, ChatMessage
from .base import (
    BaseChatBackend,
    BaseChatConfig,
    BaseChatConfigSettingsDict,
    BaseConfigSettingsDict,
    BaseEmbeddingBackend,
    BaseEmbeddingConfig,
    BaseEmbeddingConfigSettingsDict,
)


class BaseLLMSettingsDict(BaseConfigSettingsDict):
    INIT_KWARGS: NotRequired[Mapping[str, Any] | None]


class LLMBackendSettingsDict(BaseLLMSettingsDict, BaseChatConfigSettingsDict):
    PROMPT_KWARGS: NotRequired[Mapping[str, Any] | None]


class LLMEmbeddingSettingsDict(BaseLLMSettingsDict, BaseEmbeddingConfigSettingsDict):
    pass


@dataclass(kw_only=True)
class LLMBackendConfigMixin:
    init_kwargs: Mapping[str, Any]

    @classmethod
    def from_settings(cls, config: BaseLLMSettingsDict, **kwargs: Any) -> Self:
        init_kwargs = config.get("INIT_KWARGS")
        if init_kwargs is None:
            init_kwargs = {}
        kwargs.setdefault("init_kwargs", init_kwargs)

        return super().from_settings(config, **kwargs)  # type: ignore


@dataclass(kw_only=True)
class LLMChatBackendConfig(
    LLMBackendConfigMixin, BaseChatConfig[LLMBackendSettingsDict]
):
    prompt_kwargs: Mapping[str, Any]

    @classmethod
    def from_settings(cls, config: LLMEmbeddingSettingsDict, **kwargs: Any) -> Self:
        prompt_kwargs = config.get("PROMPT_KWARGS")
        if prompt_kwargs is None:
            prompt_kwargs = {}
        kwargs.setdefault("prompt_kwargs", prompt_kwargs)

        return super().from_settings(config, **kwargs)


@dataclass(kw_only=True)
class LLMEmbeddingBackendConfig(
    LLMBackendConfigMixin, BaseEmbeddingConfig[LLMEmbeddingSettingsDict]
):
    @classmethod
    def from_settings(cls, config: LLMEmbeddingSettingsDict, **kwargs: Any) -> Self:
        init_kwargs = config.get("INIT_KWARGS")
        if init_kwargs is None:
            init_kwargs = {}
        kwargs.setdefault("init_kwargs", init_kwargs)

        return super().from_settings(config, **kwargs)


class LLMChatBackend(BaseChatBackend[LLMChatBackendConfig]):
    config: LLMChatBackendConfig
    config_cls = LLMChatBackendConfig

    def chat(
        self, *, messages: Sequence[ChatMessage], stream: bool = False, **kwargs
    ) -> AIResponse:
        if stream:
            raise NotImplementedError(
                "Streaming chat is not currently supported by this backend."
            )

        model = self._get_llm_chat_model()
        full_prompt = os.linesep.join([message["content"] for message in messages])
        text_response = model.prompt(full_prompt, **self._get_prompt_kwargs()).text()
        return AIResponse(choices=[text_response])

    def _get_prompt_kwargs(self, **prompt_kwargs: Any) -> Mapping[str, Any]:
        prompt_kwargs = {}
        if self.config.prompt_kwargs is not None:
            prompt_kwargs.update(self.config.prompt_kwargs)
        return prompt_kwargs

    def _get_llm_chat_model(self) -> llm.Model:
        model = llm.get_model(self.config.model_id)
        if self.config.init_kwargs is not None:
            for config_key, config_val in self.config.init_kwargs.items():
                setattr(model, config_key, config_val)
        return model


class LLMEmbeddingBackend(BaseEmbeddingBackend[LLMEmbeddingBackendConfig]):
    config: LLMEmbeddingBackendConfig
    config_cls = LLMEmbeddingBackendConfig

    def _get_llm_embedding_model(self) -> llm.EmbeddingModel:
        model = llm.get_embedding_model(self.config.model_id)
        if self.config.init_kwargs is not None:
            for config_key, config_val in self.config.init_kwargs.items():
                setattr(model, config_key, config_val)
        return model

    def embed(self, inputs: Iterable[str]) -> Iterator[list[float]]:
        model = self._get_llm_embedding_model()
        yield from model.embed_multi(inputs)



import logging
from collections.abc import Generator, Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np

from wagtail_vector_index.storage.base import (
    Document,
    StorageProvider,
    StorageVectorIndexMixinProtocol,
)

logger = logging.Logger(__name__)


@dataclass
class ProviderConfig: ...


if TYPE_CHECKING:
    MixinBase = StorageVectorIndexMixinProtocol["NumpyStorageProvider"]
else:
    MixinBase = object


class NumpyIndexMixin(MixinBase):
    def rebuild_index(self) -> None:
        self.get_documents()

    def upsert(self, *, documents: Iterable[Document]) -> None:
        pass

    def delete(self, *, document_ids: Sequence[str]) -> None:
        pass

    def get_similar_documents(
        self, query_vector: Sequence[float], *, limit: int = 5
    ) -> Generator[Document, None, None]:
        similarities = []
        for document in self.get_documents():
            cosine_similarity = (
                np.dot(query_vector, document.vector)
                / np.linalg.norm(query_vector)
                * np.linalg.norm(document.vector)
            )
            similarities.append((cosine_similarity, document))

        sorted_similarities = sorted(
            similarities, key=lambda pair: pair[0], reverse=True
        )
        for document in [pair[1] for pair in sorted_similarities][:limit]:
            yield document


class NumpyStorageProvider(StorageProvider[ProviderConfig, NumpyIndexMixin]):
    config_class = ProviderConfig
    index_mixin = NumpyIndexMixin



from collections.abc import Mapping
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from .registry import registry  # noqa: F401

if TYPE_CHECKING:
    from wagtail_vector_index.storage.base import StorageProvider


class InvalidStorageProviderError(ImproperlyConfigured):
    pass


def get_storage_provider_config() -> Mapping:
    try:
        return settings.WAGTAIL_VECTOR_INDEX_STORAGE_PROVIDERS
    except AttributeError:
        return {
            "default": {
                "STORAGE_PROVIDER": "wagtail_vector_index.storage.numpy.NumpyStorageProvider",
            }
        }


def get_storage_provider(alias: str) -> "StorageProvider":
    provider_config = get_storage_provider_config()

    try:
        config = provider_config[alias]
    except KeyError as e:
        raise InvalidStorageProviderError(
            f"No storage provider with alias '{alias}': {e}"
        ) from e

    try:
        imported = import_string(config["STORAGE_PROVIDER"])
    except ImportError as e:
        raise InvalidStorageProviderError(
            f"Couldn't import storage provider {config['STORAGE_PROVIDER']}: {e}"
        ) from e

    params = config.copy()
    params.pop("STORAGE_PROVIDER")

    return imported(params)



from collections.abc import Mapping
from typing import Required, TypedDict, cast

from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from ..text_splitting.langchain import LangchainRecursiveCharacterTextSplitter
from ..text_splitting.naive import NaiveTextSplitterCalculator
from .base import (
    BaseBackend,
    BaseChatBackend,
    BaseConfigSettingsDict,
    BaseEmbeddingBackend,
    BaseEmbeddingConfigSettingsDict,
)


class InvalidAIBackendError(ImproperlyConfigured):
    def __init__(self, alias):
        super().__init__(f"Invalid AI backend: {alias}")


class BackendSettingsDict(TypedDict):
    CLASS: Required[str]
    CONFIG: Required[BaseConfigSettingsDict]


class ChatBackendSettingsDict(BackendSettingsDict):
    pass


class EmbeddingBackendSettingsDict(BackendSettingsDict):
    CONFIG: Required[BaseEmbeddingConfigSettingsDict]


def _validate_backend_settings(
    *, settings: BackendSettingsDict, backend_id: str
) -> None:
    if "CONFIG" not in settings:
        raise ImproperlyConfigured(
            f'AI backend settings for "{backend_id}": Missing "CONFIG".'
        )
    if not isinstance(settings["CONFIG"], Mapping):
        raise ImproperlyConfigured(
            f'AI backend settings for "{backend_id}": "CONFIG" is not a Mapping.'
        )
    if "MODEL_ID" not in settings["CONFIG"]:
        raise ImproperlyConfigured(
            f'AI backend settings for "{backend_id}": "MODEL_ID" is missing in "CONFIG".'
        )


def _get_default_text_splitter_class() -> type[LangchainRecursiveCharacterTextSplitter]:
    return LangchainRecursiveCharacterTextSplitter


def _get_default_text_splitter_length_class() -> type[NaiveTextSplitterCalculator]:
    return NaiveTextSplitterCalculator


def _get_backend(*, backend_dict: BackendSettingsDict, backend_id: str) -> BaseBackend:
    if "CLASS" not in backend_dict:
        raise ImproperlyConfigured(
            f'"AI backend "{backend_id}" settings: "CLASS" is missing in the configuration.'
        )

    try:
        backend_cls = cast(type[BaseBackend], import_string(backend_dict["CLASS"]))
    except ImportError as e:
        raise InvalidAIBackendError(
            f'"AI backend "{backend_id}" settings: "CLASS" ("{backend_dict["CLASS"]}") is not importable.'
        ) from e

    _validate_backend_settings(settings=backend_dict, backend_id=backend_id)

    backend_settings = backend_dict["CONFIG"]
    config = backend_cls.config_cls.from_settings(
        backend_settings,
    )

    return backend_cls(config=config)


def get_chat_backend(
    *, backend_dict: ChatBackendSettingsDict, backend_id: str
) -> BaseChatBackend:
    return cast(
        BaseChatBackend, _get_backend(backend_dict=backend_dict, backend_id=backend_id)
    )


def get_embedding_backend(
    *, backend_dict: EmbeddingBackendSettingsDict, backend_id: str
) -> BaseEmbeddingBackend:
    return cast(
        BaseEmbeddingBackend,
        _get_backend(
            backend_dict=backend_dict,  # type: ignore
            backend_id=backend_id,
        ),
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .base import VectorIndex


class VectorIndexRegistry:
    """A registry to keep track of all the VectorIndex classes that have been registered."""

    def __init__(self):
        self._registry: dict[str, "VectorIndex"] = {}

    def register_index(self, index: "VectorIndex"):
        self._registry[type(index).__name__] = index

    def __getitem__(self, key: str) -> "VectorIndex":
        return self._registry[key]

    def __iter__(self):
        return iter(self._registry.items())


registry = VectorIndexRegistry()
